#include <ttg.h>
#include <cmath>
#include <chrono>
#include <vector>
// needed for serialization of std::vector
#include <ttg/serialization/std/vector.h>


/* A key representing the step and position (x)
 * of a task computing a cell */
struct Key2 {
  int step;
  int x;
  // equality comparison
  bool operator==(const Key2& other) const {
    return other.step == step && other.x == x;
  }
  // hash function
  size_t hash() const {
    return (((size_t)step) << 32) ^ x;
  }
};


// Make sure our key is printable (optional)
namespace std
{
  std::ostream &operator<<(std::ostream &os,
                           const Key2 &key)
  {
    os << "{" << key.x << ", "
       << key.step << "}";
    return os;
  }
} // namespace std

static
char* getCmdOption(char ** begin, char ** end, const std::string & option);
static
bool getBoolOption(char ** begin, char ** end, const std::string & option);

using eltT = double;
int CELLSIZE = 15;
using cellT = std::vector<eltT>;
int BORDERSIZE = 4;
using borderT = std::vector<eltT>;


/* Update a cell */
static
void updateCell(const borderT &L, cellT &cell, const borderT &R) {
  // copy the cell and update in place
  cellT oldCell(cell.begin(), cell.end());
  for(int i = 0; i < CELLSIZE; i++) {
    eltT sum = 0.0;
    for(int j = -BORDERSIZE; j <= BORDERSIZE; j++) {
      if(i+j<0) { sum += L[BORDERSIZE+(i+j)] * std::pow(0.5, std::abs(j)); }
      else if(i+j < CELLSIZE) { sum += oldCell[i+j] * std::pow(0.5, std::abs(j)); }
      else { sum += R[i+j-CELLSIZE] * std::pow(0.5, std::abs(j)); }
    }
    cell[i] = sum/(2*BORDERSIZE+1);
  }
}



int main(int argc, char **argv)
{
    int nbSteps = 32;
    int nbCells = 16;
    int ncores = -1; // number of cores by default
    bool print_results = false;
    bool print_dot = false;


    /* Parse arguments */
    char* opt = nullptr;
    if ((opt = getCmdOption(argv+1, argv+argc, "-s"))) {
      nbSteps = atoi(opt);
    }

    if ((opt = getCmdOption(argv+1, argv+argc, "-n"))) {
      nbCells = atoi(opt);
    }

    if ((opt = getCmdOption(argv+1, argv+argc, "-b"))) {
      BORDERSIZE = atoi(opt);
    }

    if ((opt = getCmdOption(argv+1, argv+argc, "-c"))) {
      CELLSIZE = atoi(opt);
    }

    if ((opt = getCmdOption(argv+1, argv+argc, "-t"))) {
      ncores = atoi(opt);
    }

    if (getBoolOption(argv+1, argv+argc, "-p")) {
      print_results = true;
    }

    if (getBoolOption(argv+1, argv+argc, "-d")) {
      print_dot = true;
    }

    ttg::initialize(0, NULL, ncores);
    if (getBoolOption(argv+1, argv+argc, "-h")) {
      /* Print the help once and exit */
      if (ttg::default_execution_context().rank() == 0) {
        std::cout << argv[0] << " [-c] [-s] [-b] [-n] [-h]" << std::endl;
        std::cout << "\t-c: Cell size (" << CELLSIZE << ")\n"
                  << "\t-s: Number of steps (" << nbSteps << ")\n"
                  << "\t-b: Border size (" << BORDERSIZE << ")\n"
                  << "\t-n: Number of Cells (" << nbCells << ")\n"
                  << "\t-t: Number of threads (" << ncores << ")\n"
                  << "\t-d: Print the DOT (" << print_dot << ")\n"
                  << "\t-h: Print help" << std::endl;
      }
      ttg::finalize();
      return 0;
    }

    /* Edges from the initialization task to the compute tasks */
    ttg::Edge<Key2, cellT> I_V("I->V");
    ttg::Edge<Key2, borderT> I_L("I->L");
    ttg::Edge<Key2, borderT> I_R("I->R");

    /* Edges from a compute task to the next cell and the neighbors */
    ttg::Edge<Key2, cellT> V_V("V->V");
    ttg::Edge<Key2, borderT> V_L("V->L");
    ttg::Edge<Key2, borderT> V_R("V->R");
    ttg::Edge<int, cellT> Result("Result");


    /* Initialization task, sending a new cell to the first task and the boundaries
     * to the neighboring cells */
    auto init = ttg::make_tt<int>([=](const int &k) {
        cellT cell(CELLSIZE); // all elements zero
        if(k == 0) {
          //cell[0] = std::pow(1.0 + BORDERSIZE, nbSteps);
          cell[0] = 1.0;
        }
        borderT myLeft;
        borderT myRight;
        myLeft.insert(myLeft.begin(), cell.begin(), cell.begin()+BORDERSIZE);
        myRight.insert(myRight.begin(), cell.end()-BORDERSIZE, cell.end());
        Key2 left_k{0, (k + nbCells - 1) % nbCells};
        Key2 next_k{0, (k)};
        Key2 right_k{0, (k + 1) % nbCells};
        // TODO: add the ttg::send calls here
    },
    // input edges
    ttg::edges(),
    // output edges
    ttg::edges(I_L, I_V, I_R),
    "Init Cells", {}, {"OLeft", "OCell", "ORight"});

    /* Update a cell and send the result to the next task (output 1), it's left neighbor (0),
     * and the right neighbor (2). If we're on the last step, send to the final task. */
    auto update = ttg::make_tt([=](const Key2 &key,
                                   const borderT &L,
                                   cellT &&cell,
                                   const borderT &R)
                          {
                            int step = key.step;
                            if (step < nbSteps) {
                              updateCell(L, cell, R);
                            }
                            if( step+1 < nbSteps ) {
                              // send to next and neighbors
                              borderT myLeft;
                              borderT myRight;
                              myLeft.insert(myLeft.begin(), cell.begin(), cell.begin()+BORDERSIZE);
                              myRight.insert(myRight.begin(), cell.end()-BORDERSIZE, cell.end());
                              Key2 left_k{step+1,  (key.x + nbCells - 1) % nbCells};
                              Key2 next_k{step+1,  (key.x)};
                              Key2 right_k{step+1, (key.x + 1) % nbCells};
                              // TODO: add the ttg::send calls here
                            } else {
                                // send to result
                                ttg::send<3>( key.x, std::move(cell));
                            }
                          },
                          // input terminals
                          ttg::edges(ttg::fuse(I_R, V_R), ttg::fuse(I_V, V_V), ttg::fuse(I_L, V_L)),
                          // output terminals
                          ttg::edges(V_L, V_V, V_R, Result),
                          "Update", {"ILeft", "ICell", "IRight"}, {"OLeft", "OCell", "ORight", "OResult"});

    /* Print the results, if requested */
    auto printResult = ttg::make_tt([=](const int &k, const cellT& cell) {
        if (!print_results) return;
        std::stringstream cellStr;
        for(auto &x: cell) {
            cellStr << x << ", ";
        }
        ttg::print("Received Cell[", k, "]: ", cellStr.str());
    }, ttg::edges(Result), ttg::edges(), "Print Result", {"IResult"}, {});

    ttg::make_graph_executable(init);

    if (print_dot && ttg::default_execution_context().rank() == 0) {
      std::cout << "==== begin dot ====\n";
      std::cout << ttg::Dot(true)(init.get()) << std::endl;
      std::cout << "==== end dot ====\n";
    }


    auto start = std::chrono::high_resolution_clock::now();
    for(int i = 0; i < nbCells; i++) {
        if (init->get_world().rank() == init->get_keymap()(i)) {
            init->invoke(i);
        }
    }
    auto end_invoke = std::chrono::high_resolution_clock::now();

    ttg::execute();
    ttg::fence(init->get_world());
    auto end_fence = std::chrono::high_resolution_clock::now();

    if (init->get_world().rank() == 0) {
      ttg::print("Invoke: ", (std::chrono::duration_cast<std::chrono::microseconds>(end_invoke - start).count()) / 1000.0, "ms;",
                 "Execute: ", (std::chrono::duration_cast<std::chrono::microseconds>(end_fence - start).count()) / 1000.0, "ms");
    }

    ttg::finalize();
    return EXIT_SUCCESS;
}

char* getCmdOption(char ** begin, char ** end, const std::string & option)
{
    char ** itr = std::find(begin, end, option);
    if (itr != end && ++itr != end)
    {
        return *itr;
    }
    return nullptr;
}

bool getBoolOption(char ** begin, char ** end, const std::string & option)
{
    char ** itr = std::find(begin, end, option);
    if (itr != end)
    {
        return true;
    }
    return false;
}

